(* runtime configuration *)
(* solyga@gmx.de, 2019-03-15 *)
IMPLEMENTATION MODULE Config;

IMPORT Str, FIO, FIOStd, Lib2, Lib3, Args, SO, Utils, ConSigs, Cfg, Log;

CONST
  mid = Cfg.midConfig;
  prg = Cfg.prg;
  ver = Cfg.ver;
  ext = Cfg.ext;
  eml = Cfg.eml;
  prt = Cfg.prt;


PROCEDURE WrInfo( VAR b: SO.TBuffer );
BEGIN
  SO.WrStr( b, prg );
  SO.WrStr( b, ' v' );
  SO.WrCard( b, ver, 0 );
  SO.WrStr( b, ' (' );
  SO.WrStr( b, ext );
  SO.WrStr( b, '): UDP File Transfer Client.' );
END WrInfo;


PROCEDURE ParseArgs();

  PROCEDURE Help();
  CONST
    f ::= FIOStd.out;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    SO.Init( b, s );
    SO.Clear( b ); SO.WrStr( b, "(* " ); WrInfo( b ); SO.WrStr( b, " *)" ); SO.WrLn( b );
    FIO.WrStr( f, s );
    SO.Clear( b ); SO.WrStr( b, "syntax = '" ); SO.WrStr( b, prg ); SO.WrStr( b, "' { option } ." ); SO.WrLn( b );
    FIO.WrStr( f, s );
    SO.Clear( b ); SO.WrStr( b, "option = '-h'      (* show this info and exit *)" ); SO.WrLn( b );
    FIO.WrStr( f, s );
    SO.Clear( b ); SO.WrStr( b, "       | '-a' adr  (* set remote IP address *)" ); SO.WrLn( b );
    FIO.WrStr( f, s );
    SO.Clear( b ); SO.WrStr( b, "       | '-p' prt  (* set remote UDP port *)" ); SO.WrLn( b );
    FIO.WrStr( f, s );
    SO.Clear( b ); SO.WrStr( b, "       | '-d'      (* show debug messages *)" ); SO.WrLn( b );
    FIO.WrStr( f, s );
    SO.Clear( b ); SO.WrStr( b, "       | '-o'      (* overwrite existing file *)" ); SO.WrLn( b );
    FIO.WrStr( f, s );
    SO.Clear( b ); SO.WrStr( b, "       | '-v'      (* be verbose *) ." ); SO.WrLn( b );
    FIO.WrStr( f, s );
    Lib3.SetReturnCode( 0 );
    HALT;
  END Help;

  PROCEDURE Error( t-: ARRAY OF CHAR );
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    SO.Init( b, s );
    SO.WrStr( b, prg );
    SO.WrStr( b, '.Config.ParseArgs.Error: ' );
    SO.WrStr( b, t );
    IF Args.len > 0 THEN
      SO.WrStr( b, ": '" );
      SO.WrStr( b, Args.lit );
      SO.WrStr( b, "'" );
    END;
    SO.WrLn( b );
    SO.WrStr( b, "Try '" );
    SO.WrStr( b, prg );
    SO.WrStr( b, " -h' for help." );
    Lib3.FatalError( s );
  END Error;

  PROCEDURE GetAddress( VAR x: Types.TLngCard );
  BEGIN
    IF ( Args.sym # Args.symArg ) OR ~ Utils.MkIP( x, Args.lit ) THEN Error( 'Invalid IP address' ) END;
  END GetAddress;

  PROCEDURE GetPort( VAR p: Types.TCard );
  VAR
    x: Types.TLngCard;
    ok: BOOLEAN;
  BEGIN
    IF Args.sym # Args.symArg THEN Error( 'Invalid port number' ) END;
    x := Str.StrToCard( Args.lit, 10, ok );
    IF ~ ok THEN Error( 'Invalid port number' ) END;
    IF x > MAX( Types.TCard ) THEN Error( 'Port number out of range' ) END;
    p := Types.TCard( x );
  END GetPort;

  PROCEDURE GetCard( VAR x: LONGCARD );
  VAR
    ok: BOOLEAN;
  BEGIN
    x := Str.StrToCard( Args.lit, 10, ok );
    IF ~ ok THEN Error( 'Invalid number' ) END;
  END GetCard;

  PROCEDURE Options();
  BEGIN
    WHILE Args.sym = Args.symOpt DO
      IF    Str.Compare( Args.lit, '?' ) = 0 THEN
        Help();
      ELSIF Str.Compare( Args.lit, 'h' ) = 0 THEN
        Help();
      ELSIF Str.Compare( Args.lit, 'd' ) = 0 THEN
        data.dbg := TRUE;
      ELSIF Str.Compare( Args.lit, 'o' ) = 0 THEN
        data.ovr := TRUE;
      ELSIF Str.Compare( Args.lit, 'v' ) = 0 THEN
        data.vrb := TRUE;
      ELSIF Str.Compare( Args.lit, 'a' ) = 0 THEN
        Args.Read();
        GetAddress( data.adr );
      ELSIF Str.Compare( Args.lit, 'p' ) = 0 THEN
        Args.Read();
        GetPort( data.prt );
      ELSE
        Error( 'Invalid option' );
      END;
      Args.Read();
    END;
  END Options;

BEGIN
  Args.Reset();
  Args.Read();
  Options();
  IF Args.sym # Args.symEOA THEN Error( 'Unexpected argument' ) END;
END ParseArgs;

PROCEDURE InitDir();
VAR
  drv: SHORTCARD;
BEGIN
  drv := FIO.GetDrive();
  FIO.GetDir( drv, data.dir );
  Str.Prepend( data.dir, ':' );
  Str.Prepend( data.dir, CHR( ORD( 'A' ) + CARDINAL(drv-1) ) );
END InitDir;

PROCEDURE Show();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;

  PROCEDURE Out();
  BEGIN
    Log.Write( mid, Log.lvlMsg, s, data.vrb );
  END Out;

  PROCEDURE ShowHdr();
  BEGIN
    SO.Clear( b ); WrInfo( b ); Out();
    SO.Clear( b ); SO.WrStr( b, 'data:' ); Out();
  END ShowHdr;

  PROCEDURE Begin( n-: ARRAY OF CHAR );
  CONST w = 3;
  BEGIN
    SO.Clear( b ); SO.WrStr( b, '  ' ); SO.WrStrAdj( b, n, -w ); SO.WrStr( b, ' : ' );
  END Begin;

  PROCEDURE ShowStr( n-: ARRAY OF CHAR; v-: ARRAY OF CHAR );
  BEGIN
    Begin( n ); SO.WrChar( b, "'" ); SO.WrStr( b, v ); SO.WrChar( b, "'" ); Out();
  END ShowStr;

  PROCEDURE ShowCard( n-: ARRAY OF CHAR; v: CARDINAL );
  CONST w = 1;
  BEGIN
    Begin( n ); SO.WrLngCard( b, v, w ); Out();
  END ShowCard;

  PROCEDURE ShowIP( n-: ARRAY OF CHAR; x: Types.TLngCard );
  BEGIN
    Begin( n ); Utils.WrIP( b, x ); Out();
  END ShowIP;

BEGIN
  SO.Init( b, s );
  ShowHdr();
  ShowStr(  'prg', data.prg );
  ShowCard( 'ver', data.ver );
  ShowStr(  'ext', data.ext );
  ShowStr(  'dir', data.dir );
  ShowCard( 'ovr', ORD( data.ovr ) );
  ShowCard( 'dbg', ORD( data.dbg ) );
  ShowCard( 'vrb', ORD( data.vrb ) );
  ShowIP(   'adr', data.adr );
  ShowCard( 'prt', data.prt );
END Show;

PROCEDURE InitModule();
BEGIN
  Log.Register( mid, 'Config' );
  Lib2.Fill( data, 0 );
  Str.Copy( data.prg, prg );
  data.ver := ver;
  data.ext := ext;
  data.adr := Utils.IP( 127, 0, 0, 1 );
  data.prt := prt;
  ParseArgs();

  InitDir();
  Show();
END InitModule;

BEGIN
  InitModule();
END Config.