IMPLEMENTATION MODULE Server;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT IO;
IMPORT Lib2, SO, Tmrs, Tmr, CliIO, MD5, Crc, UFTP, Files, Utils, Cfg, Log;

CONST
  mid = Cfg.midServer;
  period = 500;  (* ms *)
  timeout = 30 * 1000;  (* ms *)
  cnt = Files.cnt;
  fglMax = UFTP.lenMaxFgm;  (* adjust here *)

TYPE
  TCnt = [ 0 .. cnt ];
  TIdx = TCnt;
  TVldIdx = TIdx[ 1 .. cnt ];

  TDnlDesc = RECORD
    i: TIdx;            (* 0 <==> free *);
    d: CliIO.TDstDesc;  (* = session ID *)
    f: Files.TFile;
    t: Tmr.TTicks;
  END;
  TDnlsDesc = RECORD
    n: TCnt;
    a: ARRAY TVldIdx OF TDnlDesc;
  END;

VAR
  dnl: TDnlsDesc;
  tmr: Tmrs.TTimer;
  ers: TErrCnts;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  IO.WrStr( 'Server.' );
  IO.WrStr( s );
  IO.WrLn();
END Message;

PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Message( s );
END Error;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Message( s );
  HALT;
END Fatal;


PROCEDURE IncErrCnt( e: TErr );
BEGIN
  Lib2.IncL( ers[e], 1 );
  Lib2.IncL( ers[eNull],1  );
END IncErrCnt;


PROCEDURE Alloc( d-: CliIO.TDstDesc; f: Files.TFile ): TIdx;
VAR
  i: TIdx;

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
    r: UFTP.TResourceID;
  BEGIN
    IF ~ ena THEN RETURN END;
    Files.GetRID( f, r );
    SO.Init( b, s );
    SO.WrStr( b, 'Alloc idx ' );
    SO.WrCard( b, i, 1 );
    SO.WrStr( b, ' dst ' );
    Utils.WrIP( b, dnl.a[i].d.a );
    SO.WrChar( b, ':' );
    SO.WrCard( b, dnl.a[i].d.p, 1 );
    SO.WrStr( b, ' rid ' );
    SO.WrLngHex( b, r, 1 );
    Message( s );
  END Show;

BEGIN
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    IF dnl.a[i].i = 0 THEN
      dnl.a[i].i := i;
      dnl.a[i].d := d;
      dnl.a[i].f := f;
      dnl.a[i].t := Tmr.Ticks();
      INC( dnl.n );
      Show();
      RETURN i;
    END;
  END;
  RETURN 0;
END Alloc;

PROCEDURE Free( i: TIdx );

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
    r: UFTP.TResourceID;
  BEGIN
    IF ~ ena THEN RETURN END;
    Files.GetRID( dnl.a[i].f, r );
    SO.Init( b, s );
    SO.WrStr( b, 'Free idx ' );
    SO.WrCard( b, i, 1 );
    SO.WrStr( b, ' dst ' );
    Utils.WrIP( b, dnl.a[i].d.a );
    SO.WrChar( b, ':' );
    SO.WrCard( b, dnl.a[i].d.p, 1 );
    SO.WrStr( b, ' rid ' );
    SO.WrLngHex( b, r, 1 );
    Message( s );
  END Show;

BEGIN
  IF i = 0 THEN RETURN END;
  IF dnl.a[i].i # i THEN Fatal( 'Free Fatal idx' ) END;
  Show();
  Files.Close( dnl.a[i].f );
  DEC( dnl.n );
  Lib2.Fill( dnl.a[i], 0 )
END Free;

PROCEDURE Find( d-: CliIO.TDstDesc ): TIdx;
VAR
  i: TIdx;
BEGIN
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    IF ( dnl.a[i].i > 0 ) & ( Lib2.Compare( dnl.a[i].d, d ) = SIZE( d ) ) THEN RETURN i END;
  END;
  RETURN 0;
END Find;


PROCEDURE OnRx( rxi: CliIO.TRxInfDesc; dat-: ARRAY OF BYTE; len: CARDINAL );
VAR
  tel: POINTER TO UFTP.TTelDesc;

  PROCEDURE SesRq();
  VAR
    i: TIdx;
    f: Files.TFile;

    PROCEDURE Reply( e: UFTP.TError );
    VAR
      rep: RECORD h: UFTP.THdrDesc; b: UFTP.TSesRpDesc; t: UFTP.TTrlDesc END;
    BEGIN
      Lib2.Fill( rep, 0 );
      rep.h.m := UFTP.mag;
      rep.h.v := UFTP.ver;
      rep.h.c := UFTP.cmdSesRp;
      rep.b.e := e;
      IF e = UFTP.errNull THEN
        Files.GetFgl( f, rep.b.f );
        Files.GetLen( f, rep.b.l );
        Files.GetMD5( f, rep.b.d );
        rep.b.r := UFTP.TReference( i );
      END;
      rep.t := Crc.Swap( Crc.UpdL( Crc.crcInit, rep, SIZE(rep) - SIZE(rep.t) ) );
      CliIO.Send( rxi.d, rep, SIZE( rep ) );
    END Reply;

  CONST
    lenMin = SIZE( UFTP.TSesRqDesc );
    lenMax = lenMin;
  VAR
    l, m: UFTP.TFgmLen;
    r: UFTP.TResourceID;
  BEGIN
    i := 0;
    f := Files.fNull;
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eSesRqLen );
      RETURN;
    END;
    l := tel^.srq.f;
    IF l = 0 THEN
      IncErrCnt( eSesRqFgl );
      RETURN;
    END;
    IF l > fglMax THEN l := fglMax END;
    i := Find( rxi.d );
    IF i > 0 THEN
      f := dnl.a[i].f;
      Files.GetRID( f, r );
      Files.GetFgl( f, m );
      IF ( tel^.srq.r = r ) & ( l = m ) THEN
        Reply( UFTP.errNull );
      ELSE
        IncErrCnt( eSesRqSes );
      END;
      RETURN;
    END;
    IF dnl.n >= cnt THEN
      IncErrCnt( eSesRqRes );
      Reply( UFTP.errRes );
      RETURN;
    END;
    f := Files.Open( tel^.srq.r, l );
    IF f = Files.fNull THEN
      IncErrCnt( eSesRqRID );
      Reply( UFTP.errRID );
      RETURN;
    END;
    i := Alloc( rxi.d, f );
    IF i = 0 THEN Fatal( 'SesRq Fatal' ) END;
    Reply( UFTP.errNull );
  END SesRq;

  PROCEDURE AppendCrc( VAR x: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;
  TYPE
    TCrcPtr = POINTER TO Crc.TCrc;
  BEGIN
    TCrcPtr( ADR(x[l]) )^ := Crc.Swap( Crc.UpdL( Crc.crcInit, x, l ) );
    RETURN l + SIZE( Crc.TCrc );
  END AppendCrc;

  PROCEDURE DatRq();
  VAR
    rep: RECORD h: UFTP.THdrDesc; b: UFTP.TDatRpDesc; t: UFTP.TTrlDesc END;
    r: UFTP.TReference;
    i: TVldIdx;
    f: Files.TFile;
    j: UFTP.TRelFgmIdx;
    l: CARDINAL;
  CONST
    lenMin = SIZE( UFTP.TDatRqDesc );
    lenMax = lenMin;
  BEGIN
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eDatRqLen );
      RETURN;
    END;
    r := tel^.drq.r;
    IF ( r < MIN( TVldIdx ) ) OR ( r > MAX( TVldIdx ) ) THEN
      IncErrCnt( eDatRqRef );
      RETURN;
    END;
    i := TVldIdx( r );
    IF ( dnl.a[i].i # i ) OR ( Lib2.Compare( dnl.a[i].d, rxi.d ) < SIZE( rxi.d ) ) THEN
      IncErrCnt( eDatRqSes );
      RETURN;
    END;
    dnl.a[i].t := Tmr.Ticks();
    f := dnl.a[i].f;
    Lib2.Fill( rep, 0 );
    rep.h.m := UFTP.mag;
    rep.h.v := UFTP.ver;
    rep.h.c := UFTP.cmdDatRp;
    FOR j := MIN( UFTP.TRelFgmIdx ) TO MAX( UFTP.TRelFgmIdx ) DO
      IF j IN tel^.drq.s THEN
        rep.b.i := tel^.drq.i + j;
        l := Files.GetFgm( f, rep.b.i, rep.b.d );
        l := AppendCrc( rep, SIZE( rep.h ) + SIZE( rep.b.i ) + l );
        CliIO.Send( rxi.d, rep, l );
      END;
    END;
  END DatRq;

  PROCEDURE Close();
  CONST
    lenMin = SIZE( UFTP.TCloseDesc );
    lenMax = lenMin;
  VAR
    r: UFTP.TReference;
    i: TVldIdx;
  BEGIN
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eCloseLen );
      RETURN;
    END;
    r := tel^.cls.r;
    IF ( r < MIN( TVldIdx ) ) OR ( r > MAX( TVldIdx ) ) THEN
      IncErrCnt( eCloseRef );
      RETURN;
    END;
    i := TVldIdx( r );
    IF ( dnl.a[i].i # i ) OR ( Lib2.Compare( dnl.a[i].d, rxi.d ) < SIZE( rxi.d ) ) THEN
      IncErrCnt( eCloseSes );
      RETURN;
    END;
    Free( i );
  END Close;

CONST
  lenMin = UFTP.lenFrmOvr;
  lenMax = UFTP.lenMaxTel;
BEGIN
  IF ( len < lenMin ) OR ( len > lenMax ) THEN IncErrCnt( eLen ); RETURN END;
  IF ~ Crc.Ok( dat, len ) THEN IncErrCnt( eCrc ); RETURN END;
  tel := ADR( dat );
  DEC( len, lenMin );
  IF tel^.hdr.m # UFTP.mag THEN IncErrCnt( eMag ); RETURN END;
  IF tel^.hdr.v # UFTP.ver THEN IncErrCnt( eVer ); RETURN END;
  CASE tel^.hdr.c OF
  | UFTP.cmdSesRq: SesRq();
  | UFTP.cmdDatRq: DatRq();
  | UFTP.cmdClose: Close();
  ELSE
    IncErrCnt( eCmd );
  END;
END OnRx;


PROCEDURE OnTmr( k: Tmrs.TTicks; r: Tmrs.TRef );
VAR
  t: Tmr.TTicks;
  i: TVldIdx;

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'OnTmr idx ' );
    SO.WrCard( b, i, 1 );
    Message( s );
  END Show;

BEGIN
  t := Tmr.Ticks();
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    IF ( dnl.a[i].i > 0 ) & ( Tmr.Diff( dnl.a[i].t, t ) > timeout ) THEN
      Show();
      Free( i );
    END;
  END;
END OnTmr;


PROCEDURE Run();
BEGIN
  CliIO.Execute();
END Run;

PROCEDURE GetErrCnts( VAR e: TErrCnts );
BEGIN
  e := ers;
END GetErrCnts;


BEGIN
  Log.Register( mid, 'Server' );
  Lib2.Fill( dnl, 0 );

  tmr := Tmrs.New();
  Tmrs.Set( tmr, period, period );
  Tmrs.SetHnd( tmr, OnTmr, 0 );

  Lib2.Fill( ers, 0 );

  CliIO.Install( OnRx );
END Server.
